/*
 * Hitori Solver in Copris
 * by Naoyuki Tamura
 * http://bach.istc.kobe-u.ac.jp/copris/puzzles/hitori/
 */
package hitori

import jp.kobe_u.copris._
import jp.kobe_u.copris.dsl._
import puzzle._

case class Hitori(m: Int, n: Int, board: Seq[Seq[String]]) extends BoardPuzzle {
  def show(sol: Set[Cell]) {
    for (i <- 0 until m) {
      for (j <- 0 until n) {
	val cell = (i,j)
	if (sol.contains(cell)) print("%3d".format(num(cell)))
	else print(" ##")
      }
      println
    }
  }
}

object Solver extends BoardPuzzleSolver[Hitori] {
  import scala.math.Ordering.Implicits._

  val name = "hitori.Solver"
  var whiteComponents: Set[Set[Cell]] = _

  def puzzleFactory(m: Int, n: Int, board: Seq[Seq[String]]) =
    Hitori(m, n, board)

  def define = {
    for (cell <- puzzle.cells)
      int('x(cell), Set(0, puzzle.num(cell)))
    for (i <- 0 until puzzle.m) {
      for (j1 <- 0 until puzzle.m; j2 <- j1+1 until puzzle.m; if puzzle.num((i,j1)) == puzzle.num((i,j2)))
	add(('x((i,j1)) === 0) || ('x((i,j2)) === 0))
    }
    for (j <- 0 until puzzle.n) {
      for (i1 <- 0 until puzzle.m; i2 <- i1+1 until puzzle.m; if puzzle.num((i1,j)) == puzzle.num((i2,j)))
	add(('x((i1,j)) === 0) || ('x((i2,j)) === 0))
    }
    for (cell <- puzzle.cells; cell1 <- puzzle.adjCells(cell))
      add(('x(cell) === 0) ==> ('x(cell1) > 0))
    // Hints
    if (puzzle.m >= 3 || puzzle.n >= 3) {
      for (cell <- puzzle.cells)
	add(('x(cell) > 0) ==> Or(puzzle.adjCells(cell).map(cell1 => 'x(cell1) > 0)))
    }
  }

  def checkSolution: Boolean = {
    val nodes = puzzle.cells.filter(cell => solution('x(cell)) > 0).toSet
    def nextCells(cell: Cell) =
      puzzle.adjCells(cell).filter(cell1 => solution('x(cell1)) > 0).toSet
    val arcs = nodes.map(cell => cell -> nextCells(cell)).toMap
    whiteComponents = getComponents(nodes, arcs)
    if (verbose >= 2)
      println("Components = " + whiteComponents.size)
    whiteComponents.size == 1
  }
  def addNegation {
    for (whites <- whiteComponents) {
      val blacks = for (cell <- whites; cell1 <- puzzle.adjCells(cell); if solution('x(cell1)) == 0)
		   yield cell1
      add(Or(whites.map(cell => 'x(cell) === 0)) || Or(blacks.map(cell => 'x(cell) > 0)))
    }
  }
  override def findFirstSolution = findIncremental(true, checkSolution, addNegation)
  override def findNextSolution = findIncremental(false, checkSolution, addNegation)

  def showSolution {
    require(whiteComponents.size == 1)
    val sol = whiteComponents.head
    if (quiet == 0) {
      println("Solution = " + sol)
      println("Size = " + sol.size)
      puzzle.show(sol)
    }
  }
}
