/*
 * Numberlink Solver in Copris
 * by Naoyuki Tamura
 * http://bach.istc.kobe-u.ac.jp/copris/puzzles/numberlink/
 */
package numberlink

import jp.kobe_u.copris._
import jp.kobe_u.copris.dsl._
import puzzle._

case class Numberlink(m: Int, n: Int, board: Seq[Seq[String]]) extends BoardPuzzle {
  def isBlank(cell: Cell) = at(cell) == "-"
  def numCells(num: Int) = cells.filter(cell => at(cell) == num.toString)
  def show(sol: Set[Seq[Cell]]) {
    val arcs = sol.flatMap(_.sliding(2).map(e => (e(0),e(1))).toSet)
    def connected(cell1: Cell, cell2: Cell) =
      arcs.contains((cell1,cell2)) || arcs.contains((cell2,cell1))
    println("." + "   ." * n)
    for (i <- 0 until m) {
      print("  ")
      for (j <- 0 until n) {
	val cell = (i,j); val cell1 = (i,j+1)
	val x = if (isBlank(cell)) "+" else at(cell)
	val e = if (isCell(cell1) && connected(cell, cell1)) "---"
		else "   "
	print((x + e).substring(0, 4))
      }
      println
      print(".")
      for (j <- 0 until n) {
	val cell = (i,j); val cell1 = (i+1,j)
	val e = if (isCell(cell1) && connected(cell, cell1)) " | "
		else "   "
	print(e + ".")
      }
      println
    }
    println
  }
}

object Solver extends BoardPuzzleSolver[Numberlink] {
  import scala.math.Ordering.Implicits._

  val name = "numberlink.Solver"

  def puzzleFactory(m: Int, n: Int, board: Seq[Seq[String]]) =
    Numberlink(m, n, board)

  def define = {
    for (cell <- puzzle.cells; cell1 <- puzzle.adjCells(cell))
      int('e(cell,cell1), 0, 1)
    for (cell <- puzzle.cells; cell1 <- puzzle.adjCells(cell); if cell < cell1)
      add('e(cell,cell1) + 'e(cell1,cell) <= 1)
    for (cell <- puzzle.cells) {
      val in = puzzle.adjCells(cell).map(cell1 => 'e(cell1,cell))
      val ot = puzzle.adjCells(cell).map(cell1 => 'e(cell,cell1))
      if (puzzle.isBlank(cell)) {
	int('io(cell), 0, 1)
	add('io(cell) === Add(in))
	add('io(cell) === Add(ot))
      } else {
	val numCells = puzzle.numCells(puzzle.num(cell))
	if (cell == numCells(0)) {
	  add(Add(in) === 0)
	  add(Add(ot) === 1)
	} else {
	  add(Add(in) === 1)
	  add(Add(ot) === 0)
	}
      }
    }
    val numbers = puzzle.numberCells.map(puzzle.num(_)).toSet
    for (cell <- puzzle.cells) {
      if (puzzle.isBlank(cell))
	int('x(cell), numbers)
      else
	int('x(cell), puzzle.num(cell))
    }
    for (cell <- puzzle.cells; cell1 <- puzzle.adjCells(cell))
      add(('e(cell,cell1) === 1) ==> ('x(cell) === 'x(cell1)))
  }

  def showSolution {
    def findPath(cell: Cell): Seq[Cell] =
      puzzle.adjCells(cell).find(cell1 => solution('e(cell,cell1)) > 0) match {
	case None => Seq(cell)
	case Some(cell1) => cell +: findPath(cell1)
      }
    val numbers = puzzle.numberCells.map(puzzle.num(_)).toSet
    val sol = for (num <- numbers; cell = puzzle.numCells(num)(0))
	      yield findPath(cell)
    if (quiet == 0) {
      println("Solution = " + sol)
      println("Size = " + sol.map(_.size).sum)
      puzzle.show(sol)
    }
  }
}
