/*
 * Shakashaka Solver in Copris
 * by Naoyuki Tamura
 * http://bach.istc.kobe-u.ac.jp/copris/puzzles/shakashaka/
 */
package shakashaka

import jp.kobe_u.copris._
import jp.kobe_u.copris.dsl._
import puzzle._

case class Shakashaka(m: Int, n: Int, board: Seq[Seq[String]]) extends BoardPuzzle {
  def isBlank(cell: Cell) = at(cell) == "-"
  def isBlack(cell: Cell) = isNumber(cell) || at(cell) == "x"
  val WH = 0; val NW = 1; val NE = 2; val SE = 3; val SW = 4; val BX = 5
  def show(sol: Map[Cell,Int]) {
    val tiles = Map(WH -> "\u25A1", NW -> "\u25E4", NE -> "\u25E5", SE -> "\u25E2", SW -> "\u25E3", BX -> "\u25A0")
    // val tiles = Map(WH -> "..", NW -> "#/", NE -> "\\#", SE -> "/#", SW -> "#\\", BX -> "##")
    for (i <- 0 until m) {
      for (j <- 0 until n) {
	val cell = (i,j)
	if (isBlank(cell)) print(tiles(sol(cell)))
	else if (isNumber(cell)) print("#" + num(cell))
	else print(tiles(BX))
      }
      println
    }
  }
}

object Solver extends BoardPuzzleSolver[Shakashaka] {
  val name = "shakashaka.Solver"

  def puzzleFactory(m: Int, n: Int, board: Seq[Seq[String]]) =
    Shakashaka(m, n, board)

  def WH = puzzle.WH
  def NW = puzzle.NW; def NE = puzzle.NE; def SE = puzzle.SE; def SW = puzzle.SW
  def BX = puzzle.BX

  def define = {
    for (i <- -1 to puzzle.m; j <- -1 to puzzle.n) {
      val cell = (i,j)
      if (! puzzle.isCell(cell) || puzzle.isBlack(cell))
	int('x(cell), BX)
      else
	int('x(cell), Set(WH,NW,NE,SE,SW))
    }
    for (cell <- puzzle.cells; if puzzle.isNumber(cell) && puzzle.num(cell) <= 4) {
      val xs = for (cell1 <- puzzle.adjCells(cell); if puzzle.isBlank(cell1))
	       yield If('x(cell1) =/= WH, Num(1), Num(0))
      add(Add(xs) === puzzle.num(cell))
    }
    def x(cell: Cell, dij: Cell) = 'x(puzzle.move(cell, dij))
    for (cell <- puzzle.cells; if puzzle.isBlank(cell)) {
      add((x(cell,( 0, 0)) === NW) ==> (
	 ((x(cell,(-1, 1)) === NW) && (x(cell,( 0, 1)) === WH)) || (x(cell,( 0, 1)) === NE)))
      add((x(cell,( 0, 0)) === NW) ==> (
	 ((x(cell,( 1,-1)) === NW) && (x(cell,( 1, 0)) === WH)) || (x(cell,( 1, 0)) === SW)))
      add((x(cell,( 0, 0)) === NE) ==> (
	 ((x(cell,(-1,-1)) === NE) && (x(cell,( 0,-1)) === WH)) || (x(cell,( 0,-1)) === NW)))
      add((x(cell,( 0, 0)) === NE) ==> (
	 ((x(cell,( 1, 1)) === NE) && (x(cell,( 1, 0)) === WH)) || (x(cell,( 1, 0)) === SE)))
      add((x(cell,( 0, 0)) === SE) ==> (
	 ((x(cell,(-1, 1)) === SE) && (x(cell,(-1, 0)) === WH)) || (x(cell,(-1, 0)) === NE)))
      add((x(cell,( 0, 0)) === SE) ==> (
	 ((x(cell,( 1,-1)) === SE) && (x(cell,( 0,-1)) === WH)) || (x(cell,( 0,-1)) === SW)))
      add((x(cell,( 0, 0)) === SW) ==> (
	 ((x(cell,(-1,-1)) === SW) && (x(cell,(-1, 0)) === WH)) || (x(cell,(-1, 0)) === NW)))
      add((x(cell,( 0, 0)) === SW) ==> (
	 ((x(cell,( 1, 1)) === SW) && (x(cell,( 0, 1)) === WH)) || (x(cell,( 0, 1)) === SE)))
    }
    def nEdge(x: Var) = (x === NW) || (x === NE) || (x === BX)
    def eEdge(x: Var) = (x === NE) || (x === SE) || (x === BX)
    def sEdge(x: Var) = (x === SE) || (x === SW) || (x === BX)
    def wEdge(x: Var) = (x === SW) || (x === NW) || (x === BX)
    for (cell <- puzzle.cells) {
      add(((x(cell,(0,0)) === WH) && sEdge(x(cell,(-1, 0)))) ==>
	  (sEdge(x(cell,(-1,-1))) || eEdge(x(cell,( 0,-1)))))
      add(((x(cell,(0,0)) === WH) && sEdge(x(cell,(-1, 0)))) ==>
	  (sEdge(x(cell,(-1, 1))) || wEdge(x(cell,( 0, 1)))))
      add(((x(cell,(0,0)) === WH) && wEdge(x(cell,( 0, 1)))) ==>
	  (wEdge(x(cell,(-1, 1))) || sEdge(x(cell,(-1, 0)))))
      add(((x(cell,(0,0)) === WH) && wEdge(x(cell,( 0, 1)))) ==>
	  (wEdge(x(cell,( 1, 1))) || nEdge(x(cell,( 1, 0)))))
      add(((x(cell,(0,0)) === WH) && nEdge(x(cell,( 1, 0)))) ==>
	  (nEdge(x(cell,( 1, 1))) || wEdge(x(cell,( 0, 1)))))
      add(((x(cell,(0,0)) === WH) && nEdge(x(cell,( 1, 0)))) ==>
	  (nEdge(x(cell,( 1,-1))) || eEdge(x(cell,( 0,-1)))))
      add(((x(cell,(0,0)) === WH) && eEdge(x(cell,( 0,-1)))) ==>
	  (eEdge(x(cell,( 1,-1))) || nEdge(x(cell,( 1, 0)))))
      add(((x(cell,(0,0)) === WH) && eEdge(x(cell,( 0,-1)))) ==>
	  (eEdge(x(cell,(-1,-1))) || sEdge(x(cell,(-1, 0)))))
    }
  }

  def showSolution {
    val sol = {
      for (cell <- puzzle.cells.toSet; if puzzle.isBlank(cell))
      yield cell -> solution('x(cell))
    }.toMap
    if (quiet == 0) {
      println("Solution = " + sol)
      println("Size = " + sol.size)
      puzzle.show(sol)
    }
  }
}
